@@ -79,9 +79,11 @@
 #define PS_VERSION			20
 #define PS_HEADER			30
 #define PS_HEADER_END		35
-#define PS_END_OF_REQUEST	40
-#define PS_HOST_HEADER		50
-#define PS_REFERER_HEADER	60
+#define PS_POST_DATA		40
+#define PS_END_OF_REQUEST	45
+#define PS_HOST_HEADER		50	/* Needs 10 slots */
+#define PS_REFERER_HEADER	60	/* Needs 10 slots */
+#define PS_LENGTH_HEADER	70	/* Needs 20 slots */
 
 #define RC_TRANSPARENT		1
 #define RC_FORCED			2
@@ -117,6 +119,7 @@ typedef struct connection_
 	int					host_header_length;
 	char				referer_header[REFERRER_SIZE];
 	int					referer_header_length;
+	int					content_length;
 	int					request_number;
 }	connection_t;
 
@@ -618,7 +621,7 @@ static uid_t getuserid(char *user)
 	}
 
 #ifdef DEBUG
-	printf("User lookup %s -> %d\n", user, uid);
+	printf("User lookup %s -> %lu\n", user, (unsigned long)uid);
 #endif
 
 	endpwent();
@@ -644,7 +647,7 @@ static uid_t getgroupid(uid_t uid)
 	}
 
 #ifdef DEBUG
-	printf("Group lookup %d -> %d\n", uid, gid);
+	printf("Group lookup %lu -> %lu\n", (unsigned long)uid, (unsigned long)gid);
 #endif
 
 	endpwent();
@@ -1337,14 +1340,25 @@ static int process_client_request(connection_t *conn, size_t last_read)
 
 		case PS_HEADER_END:
 			if (c == '\r')					conn->parse_state = PS_HEADER_END + 1;
-			else if (c == '\n')				conn->parse_state = PS_END_OF_REQUEST;
+			else if (c == '\n') {
+				if (conn->content_length > 0)
+					conn->parse_state = PS_POST_DATA;
+				else
+					conn->parse_state = PS_END_OF_REQUEST;
+			}
 			else if (c == 'H' || c == 'h')	conn->parse_state = PS_HOST_HEADER;
 			else if (c == 'R' || c == 'r')	conn->parse_state = PS_REFERER_HEADER;
+			else if (c == 'C' || c == 'c')	conn->parse_state = PS_LENGTH_HEADER;
 			else							conn->parse_state = PS_HEADER;
 			break;
 
 		case PS_HEADER_END + 1:
-			if (c == '\n')					conn->parse_state = PS_END_OF_REQUEST;
+			if (c == '\n') {
+				if (conn->content_length > 0)
+					conn->parse_state = PS_POST_DATA;
+				else
+					conn->parse_state = PS_END_OF_REQUEST;
+			}
 			else							conn->parse_state = PS_HEADER;
 			break;
 
@@ -1454,6 +1468,130 @@ static int process_client_request(connection_t *conn, size_t last_read)
 			else if (conn->referer_header_length < (sizeof(conn->referer_header) - 1))
 				conn->referer_header[conn->referer_header_length++] = c;
 			break;
+
+		case PS_LENGTH_HEADER:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'O' || c == 'o')	conn->parse_state = PS_LENGTH_HEADER + 1;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 1:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'N' || c == 'n')	conn->parse_state = PS_LENGTH_HEADER + 2;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 2:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'T' || c == 't')	conn->parse_state = PS_LENGTH_HEADER + 3;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 3:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'E' || c == 'e')	conn->parse_state = PS_LENGTH_HEADER + 4;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 4:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'N' || c == 'n')	conn->parse_state = PS_LENGTH_HEADER + 5;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 5:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'T' || c == 't')	conn->parse_state = PS_LENGTH_HEADER + 6;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 6:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == '-')				conn->parse_state = PS_LENGTH_HEADER + 7;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 7:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'L' || c == 'l')	conn->parse_state = PS_LENGTH_HEADER + 8;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 8:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'E' || c == 'e')	conn->parse_state = PS_LENGTH_HEADER + 9;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 9:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'N' || c == 'n')	conn->parse_state = PS_LENGTH_HEADER + 10;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 10:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'G' || c == 'g')	conn->parse_state = PS_LENGTH_HEADER + 11;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 11:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'T' || c == 't')	conn->parse_state = PS_LENGTH_HEADER + 12;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 12:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == 'H' || c == 'h')	conn->parse_state = PS_LENGTH_HEADER + 13;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 13:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c == ':')				conn->parse_state = PS_LENGTH_HEADER + 14;
+			else							conn->parse_state = PS_HEADER;
+			break;
+
+		case PS_LENGTH_HEADER + 14:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c != ' ') {			conn->parse_state = PS_LENGTH_HEADER + 15;
+				if (c >= '0' && c <= '9')
+					conn->content_length = c - '0';
+				else
+					conn->content_length = 0;
+			}
+			break;
+
+		case PS_LENGTH_HEADER + 15:
+			if (c == '\r')					conn->parse_state = PS_HEADER + 1;
+			else if (c == '\n')				conn->parse_state = PS_HEADER_END;
+			else if (c >= '0' && c <= '9') {
+				conn->content_length *= 10;
+				conn->content_length += c - '0';
+			}
+			break;
+
+		case PS_POST_DATA:
+			if (conn->content_length <= 1)	conn->parse_state = PS_END_OF_REQUEST;
+			else
+				--conn->content_length;
+			break;
 		}
 
 		/*
@@ -1574,7 +1712,8 @@ static int process_client_request(connection_t *conn, size_t last_read)
 				memcpy(&request_buffer[send_size], conn->host_header, conn->host_header_length);
 				send_size += conn->host_header_length;
 				conn->url_end_offset += conn->host_header_length;
-				if (ntohs(conn->dest_addr.sin_port) != 80)
+				if (ntohs(conn->dest_addr.sin_port) != 80 &&
+				    strchr(conn->host_header, ':') == NULL)
 				{
 					sprintf(request_port, ":%u", ntohs(conn->dest_addr.sin_port));
 					memcpy(&request_buffer[send_size], request_port, strlen(request_port));
@@ -1699,6 +1838,7 @@ static int process_client_request(connection_t *conn, size_t last_read)
 			conn->host_header[0] = '\0';
 			conn->host_header_length = 0;
 			conn->referer_header_length = 0;
+			conn->content_length = 0;
 		}
 	}
 
@@ -1773,7 +1913,7 @@ static void write_pid(char *prog)
 	sprintf(filename, "%s%s.pid", _PATH_VARRUN, prog);
 	if ((fp = fopen(filename, "w")) != NULL)
 	{
-		fprintf(fp, "%u\n", getpid());
+		fprintf(fp, "%lu\n", (unsigned long)getpid());
 		fclose(fp);
 	}
 }
